
import hysteresis.curve as hys
import numpy as np
import pytest

def makeCurve(f):
    "lamda function that creates all curves"
    xdata = np.linspace(0,4,1001)
    y = f(xdata)
    xy = np.column_stack([xdata, y])
    Curve = hys.Curve(xy)   
    return Curve    

def function1(x):
    return x + 2

def function2(x):
    return x**3 - 3*x**2 + 3

def function3(x):
    return 3 + (np.e)**x  
    
def test_len():
    curve = makeCurve(function2)
    length = len(curve)
    assert length == 1001

# =============================================================================
# 
# =============================================================================

def test_get_Operand_scalar():
    assert hys._getOperand(2) == 2

def test_get_Operand_flat_array():
    arrayIn = np.linspace(0,4,101)
    assert np.all(hys._getOperand(arrayIn) == arrayIn)

def test_get_Operand_2D_array():
    xdata = np.linspace(0,4,1001)
    y = function3(xdata)
    xy = np.column_stack([xdata, y])
    assert np.all(hys._getOperand(xy) == y)

def test_get_Operand_2D_array_failure():
    with pytest.raises(Exception):
        xdata = np.linspace(0,4,1001)
        y = function3(xdata)
        xy = np.column_stack([xdata, y,y])
        hys._getOperand(xy)

# =============================================================================
# 
# =============================================================================

def test_multiplication_scalar():
    curve = makeCurve(function2)
    newCurve = curve*8
    assert np.all(newCurve.xy[:,1] == curve.xy[:,1]*8)

def test_multiplication_Array():
    curve = makeCurve(function2)
    newCurve = curve*curve.xy
    assert np.all(newCurve.xy[:,1] == curve.xy[:,1]**2)

def test_multiplication_Curve():
    curve = makeCurve(function3)
    newCurve = curve*curve
    assert np.all(newCurve.xy[:,1] == curve.xy[:,1]**2)
# =============================================================================
# 
# =============================================================================

def test_addition_scalar():
    curve = makeCurve(function2)
    newCurve = curve+8
    assert np.all(newCurve.xy[:,1] == curve.xy[:,1]+8)

def test_addition_Array():
    curve = makeCurve(function2)
    newCurve = curve + curve.xy
    assert np.all(newCurve.xy[:,1] == curve.xy[:,1]*2)

def test_addition_Curve():
    curve = makeCurve(function3)
    newCurve = curve+curve
    assert np.all(newCurve.xy[:,1] == curve.xy[:,1]*2)

# =============================================================================
# 
# =============================================================================

def test_subtraction_scalar_L():
    curve = makeCurve(function2)
    newCurve = curve-8
    assert np.all(newCurve.xy[:,1] == curve.xy[:,1]-8)

def test_subtraction_scalar_R():
    curve = makeCurve(function2)
    newCurve = 8 - curve
    assert np.all(newCurve.xy[:,1] == 8 - curve.xy[:,1])
    
    
def test_subtraction_Array_L():
    curve = makeCurve(function2)
    newCurve = curve - curve.xy
    assert np.all(newCurve.xy[:,1] == 0)
    
def test_subtraction_Array_R():
    curve = makeCurve(function2)
    newCurve =  curve.xy - curve 
    assert np.all(newCurve.xy[:,1] == 0)
    
    
def test_subtraction_Curve():
    curve = makeCurve(function3)
    newCurve = curve-curve
    assert np.all(newCurve.xy[:,1] == 0)

# =============================================================================
# 
# =============================================================================

def test_division_scalar_L():
    curve = makeCurve(function2)
    newCurve = curve/8
    assert np.all(newCurve.xy[:,1] == curve.xy[:,1]/8)

def test_division_scalar_R():
    curve = makeCurve(function2)
    newCurve = 8 / curve
    assert np.all(newCurve.xy[:,1] == 8 / curve.xy[:,1])
    
    
def test_division_Array_L():
    curve = makeCurve(function2)
    newCurve = curve / curve.xy
    assert np.all(newCurve.xy[:,1] == 1)
    
def test_division_Array_R():
    curve = makeCurve(function2)
    newCurve =  curve.xy / curve 
    assert np.all(newCurve.xy[:,1] == 1)
    
    
def test_division_Curve():
    curve = makeCurve(function3)
    newCurve = curve/curve
    assert np.all(newCurve.xy[:,1] == 1)
    


# curve = makeCurve(function2)
# test = np.array(curve)



# test_subtraction_scalar_L()
# test_subtraction_scalar_R()
# test_subtraction_Array_L()
# # test_subtraction_Array_R()
# test_subtraction_Curve()


import numpy as np
from .curve import Hysteresis, SimpleCurve, MonotonicCurve
from .baseFuncs import (concatenateHys, concatenate, _linInterp, 
                        _linInterpSample, _getNsamples)
       
# =============================================================================
# concatenate and resample
# =============================================================================

def _resampleSimpleCurve(curve, Nsamples):
    x = curve.xy[:,0]
    y = curve.xy[:,1]
    
    # If the simple cycle has subcycles, resample at the level of those.
    if curve.subCycles:
        outputSubcycles = [None]*curve.NsubCycles
        for ii, subcycle in enumerate(curve.subCycles):
            outputSubcycles[ii] = resample(subcycle, Nsamples)    
            
        tmpOut      = concatenate(outputSubcycles, outputClass = SimpleCurve)
        tmpProps    = curve._getStatePropreties()
        return SimpleCurve(tmpOut.xy, *tmpProps)
                    
    # Otherwise, resample the curve directly
    else:
        return SimpleCurve(_linInterp(x,y, Nsamples))  


def _resampleHysteresis(curve, Nsamples):
    outputCycles = [None]*curve.NCycles
    for ii, cycle in enumerate(curve.cycles):
        outputCycles[ii] = resample(cycle, Nsamples)      
    
    tmpOut = concatenateHys(outputCycles)    # If curve
    tmpProps = curve._getStatePropreties()
    return Hysteresis(tmpOut.xy,*tmpProps)      

def _resampleMonotonicCurve(curve, Nsamples):
    x = curve.xy[:,0]
    y = curve.xy[:,1]
    return MonotonicCurve(_linInterp(x,y, Nsamples))     

def _resampleArray(curve, Nsamples):
    x = curve[:,0]
    y = curve[:,1]
    return _linInterp(x,y, Nsamples)  



def resample(curve, Nsamples, **kwargs):
    """
    Creates a new curve object that has a different number of sample points 
    between the start and end of each Cycle object.
    The curve resampled will be one level down, i.e. Hystereses will their
    simpleCycles resampled, and SimpleCycles will have their Monotonic Curves
    resampled
    
    The number of points is specified by the user, and intermediate points will
    be evenly spaced between the cycles start and end points.
    Linear interpolation is used to find the y value of intermediate points.
    In the case of a Hysteresis, every SimpleCycle curve will be resampled with
    an amount of points equal to Nsamples.
    
    In the case of a SimpleCycle, Monotonic curves are resampled if they are set.

    Parameters
    ----------
    curve : Hysteresis, Monotonic Cycle, or numpy array
        The curve to be resampled. If a numpy array is provided, it must be 
        increasing to be resampled properly.
    Nsamples : 
        The number of samples for the new curve.
        
    Returns
    -------
    curve : Hysteresis, SimpleCycle, 
        An object of the input type with a number of points equal to Nsamples

    """  
        
    # We recursively resample by calling the function again for sub-cycle objects
    
    # if the curve is a SimpleCycle
    if isinstance(curve, SimpleCurve):
        output = _resampleSimpleCurve(curve, Nsamples)
       
    # if the curve is a hysteresis, we recursively create a series of Cycles
    elif isinstance(curve, Hysteresis):
        output = _resampleHysteresis(curve, Nsamples)
        
    # if the curve is a Monotonic Cycle
    elif isinstance(curve, MonotonicCurve):
        output = _resampleMonotonicCurve(curve, Nsamples)
    
    # if it is a np array
    elif isinstance(curve, np.ndarray):
        output = _resampleArray(curve, Nsamples)
        
    else:
        output = None
        
    return output




def _resampledxSimpleCurve(curve, targetdx):
    if curve.subCycles: # if subsycles are set, resample those
        outputSubcycles = [None]*curve.NsubCycles
        for ii, subcycle in enumerate(curve.subCycles):
            outputSubcycles[ii] = resampleDx(subcycle, targetdx)        
        tmpOut      =  concatenate(outputSubcycles, outputClass = SimpleCurve)
        tmpProps    = curve._getStatePropreties()
        return SimpleCurve(tmpOut.xy, *tmpProps)        
    
    else: # if subsycles aren't set, resample those
        x = curve.xy[:,0]
        dxNet = x[-1] - x[0]
        Nsamples = _getNsamples(targetdx, dxNet)
        return resample(curve, Nsamples)     


def _resampledxHysteresis(curve, targetdx):
    outputCycles = [None]*curve.NCycles
    for ii, cycle in enumerate(curve.cycles):
        outputCycles[ii] = resampleDx(cycle, targetdx)       
    tmpOut = concatenateHys(outputCycles)    # If curve
    tmpProps = curve._getStatePropreties()
    return Hysteresis(tmpOut.xy,*tmpProps)         

def _resampledxMonotonicCurve(curve, targetdx):
    x = curve.xy[:,0]
    dxNet = x[-1] - x[0]
    Nsamples = _getNsamples(targetdx, dxNet)
    return resample(curve, Nsamples)      

def _resampledxArray(curve, targetdx):
    x = curve[:,0]
    dxNet = x[-1] - x[0]
    Nsamples = _getNsamples(targetdx, dxNet)        
    return resample(curve, Nsamples)



def resampleDx(curve, targetdx):
    """
    Creates a new curve object where the distance between sample points
    is approximately dx. The number of samples betweent the start and 
    end point of each cycle is calculated so that the is as close to dx 
    as possible. Note that distance isn't precisely dx.
    

    For each curve, a number of points will be chosen based on some target
    displacement value in the x direction. If the value overshoots, 
    the last point will instead be used.
    For example, if the start = 2.5, end = 7.5 and dx = 2, points will be 
    placed at 2.5, 4.5, 6.5, 7.5 
    
    Linear interpolation is used to find the y value of intermediate points.
    In the case of a Hysteresis, every SimpleCurve will be resampled with
    an amount of points equal to Nsamples.
    In the case of a SimpleCurve, each monotonicCurve will be resampled.

    Parameters
    ----------
    curve : Hysteresis, Monotonic Cycle, or numpy array
        The curve to be resampled.

    Targetdx : float
        The target distance to use for sample points.
        
    Returns
    -------
    curve :
        An object of the input type, with points approximately at Targetdx.

    """  

    # the logic in these functions is getting a little nasty, in the future
    # consider pulling these out into class methods.
    if isinstance(curve, SimpleCurve):
        output = _resampledxSimpleCurve(curve, targetdx) 
                
    elif isinstance(curve, Hysteresis):
        output = _resampledxHysteresis(curve, targetdx) 
       
    # if the curve is a MonotonicCurve Cycle
    elif isinstance(curve, MonotonicCurve):
        output = _resampledxMonotonicCurve(curve, targetdx) 
    
    # if it is a np array
    elif isinstance(curve, np.ndarray):
        output = _resampledxArray(curve, targetdx) 
        
    return output




def _parseRegions(regions):
    """
    Ensures each region either has only two items, and
    is specified in a list of lists.
    """
    if len(regions) == 2 and isinstance(regions[0], float):
        regions = [regions]
    
    for item in regions:
        if len(item) !=2:
            raise Exception
    return regions

def _getBreakPoints(regions):
    breakPoints = set()
    for region in regions:
        breakPoints.update(set(region))
    return list(breakPoints)


def _getOutsideRegions(regions):
    """
    Gets the regions outside of the specified region
    """
    xmin = 0
    xmax = 1
    
    breakPoints = _getBreakPoints(regions)    
    
    outsideRegions = []
    isInside = []
    if xmin < breakPoints[0]:
        outsideRegions.append([xmin, breakPoints[0]])
        isInside.append(False)
    
    Nregions = len(regions)
    for ii in range(Nregions):
        outsideRegions.append(regions[ii])
        isInside.append(True)
        if ii != Nregions - 1: # check if we aren't out of bounds.
            outsideRegions.append([regions[ii][1], regions[ii+1][0]])
            isInside.append(False)
            
    if  breakPoints[-1] < xmax:
        outsideRegions.append([breakPoints[-1], xmax])        
        isInside.append(False)
        
    return outsideRegions, isInside




def _resampleRegionSimpleCurve(curve, Nsamples, regions):
    if curve.subCycles: # if subsycles are set, resample those
        outputSubcycles = [None]*curve.NsubCycles
        for ii, subcycle in enumerate(curve.subCycles):
            outputSubcycles[ii] = resampleRegion(subcycle, Nsamples, regions)
            
        tmpOut      =  concatenate(outputSubcycles, outputClass = SimpleCurve)
        tmpProps    = curve._getStatePropreties()
        return SimpleCurve(tmpOut.xy, *tmpProps)        
    
    else: # if subsycles aren't set, resample at level of xy
    
        x = curve.xy[:,0]
        y = curve.xy[:,1]
        xy = np.column_stack([x, y])
        return SimpleCurve(resampleRegion(xy, Nsamples, regions))    


def _resampleRegionHysteresis(curve, Nsamples, regions):
    outputCycles = [None]*curve.NCycles
    for ii, cycle in enumerate(curve.cycles):
        outputCycles[ii] = resampleRegion(cycle, Nsamples, regions)
    tmpOut = concatenateHys(outputCycles)
    tmpProps = curve._getStatePropreties()
    return Hysteresis(tmpOut.xy,*tmpProps)        

def _resampleRegionMonotonicCurve(curve, Nsamples, regions):
    x = curve.xy[:,0]
    y = curve.xy[:,1]
    xy = np.column_stack([x, y])
    return  MonotonicCurve(resampleRegion(xy, Nsamples, regions))        



def _resampleRegionArray(curve, Nsamples, regions):
    
    allRegions, isInside = _getOutsideRegions(regions) # these get left alone  
    x = curve[:,0]
    y = curve[:,1] 

    xSamples = []
    ySamples = []
    
    allRegions = np.array(allRegions)*(x[-1] - x[0]) + x[0]
    
    if x[0] <= x[-1]:
        isLeftRight = True
    else:
        isLeftRight = False
    
    for ii, region in enumerate(allRegions):
        if isInside[ii]:
            xtmp = np.linspace(region[0], region[1], Nsamples)
            ytmp = _linInterpSample(x, y, xtmp)[:,1]
    
        else:
            if isLeftRight:
                inds = np.where((region[0] < x) *  (x < region[1]))
            else:
                inds = np.where((x < region[0]) *  (region[1] < x))

            xtmp = x[inds]
            ytmp = y[inds]
        
        xSamples.append(xtmp)
        ySamples.append(ytmp)
        
    xSamples = np.concatenate(xSamples)
    ySamples = np.concatenate(ySamples)
    return np.column_stack([xSamples, ySamples])

def resampleRegion(curve, Nsamples, regions = None):
    """
    Resamples only parts of an input a curve. The part of the
    curve that is resampled is defined using the regions variable.
    The regions is a list of lists, defining a start and end point
    of the curve in terms of a parameter varying between 0 and 1.
    
    Each region is resampled using Nsamples

    Parameters
    ----------
    curve : Hysteresis Curve, or numpy array
        The input curve to be resampled. Can be either a hystersis curve,
        or a numpy array. If a numpy array is provided, it must not change
        direction to be resampled properly.
    region : list of lists, or list
        A lists of regions, written in terms of a parameter that ranges
        from 0 at the start of the curve to 1 at the end of the curve.
        Regions are either lists of lists, or a single list if the number 
        of regions is equal to one.
        
        Regions must be sorted from left to right and have no overlaps, i.e. 
        [[0, 0.2], [0.1, 0.2]] is invalid
        [[0, 0.1], [0.8, 0.9], [0.4, 0.5]] is invalid
        
    Nsample : int
        The number of sample to use for each region.

    Returns
    -------
    None.

    """
    if not regions:
        regions = [[0, 0.1], [0.9, 1]]
    
    regions = _parseRegions(regions) # these get interpolated
        
    if isinstance(curve, SimpleCurve):
        output = _resampleRegionSimpleCurve(curve, Nsamples,regions)
    
    elif isinstance(curve, Hysteresis):
        output = _resampleRegionHysteresis(curve, Nsamples, regions)
    
    elif isinstance(curve, MonotonicCurve):
        output = _resampleRegionMonotonicCurve(curve, Nsamples, regions)
    
    # if it is a np array
    elif isinstance(curve, np.ndarray):
        output = _resampleRegionArray(curve, Nsamples, regions)
        
    return output




from .defaultDataFuncs import (defaultAreaFunction, defaultSlopeFunction, 
                                defaultSampleFunction, defaultCombineDiff,
                                defaultLengthFunction)

from .defaultPlotFuncs import initializeFig, defaultPlotFunction, defaultShowCycles



# =============================================================================
# Curve objects
# =============================================================================



class HYSTERESIS_ENVIRONMENT:
    
    def __init__(self):
        """
        Contains the standard functions needed by the hysteresis analysis.
        """
        
        self.fslope = defaultSlopeFunction
        self.fArea = defaultAreaFunction
        self.flength = defaultLengthFunction
        
        self.finit = initializeFig
        self.fplot = defaultPlotFunction
        self.fcycles = defaultShowCycles
        
        self.fSample = defaultSampleFunction
        self.fCombineDiff = defaultCombineDiff
        
    def restart(self):
        self.__init__()


environment = HYSTERESIS_ENVIRONMENT()




import numpy as np
from scipy.interpolate import interp1d
from hysteresis import data
from .data import linearInterpolation
from .curve import Hysteresis, SimpleCurve
from .baseFuncs import concatenate



 
# =============================================================================
# Find the backbone
# =============================================================================

def _LPparser(LPsteps):
    """
    Used to find the index of, including the first point (starting at zero),
    and first peak
    """
    
    # InputParser - Decides what to do with the variable LoadProtcol
    # if you get an interger use that for all cycles
    
    if len(LPsteps) ==0:
        Indexes = []
    
    # if type(LPsteps) == int:
    else:
        # shift = LPsteps[0] - 1
        shift = 1
        Indexes = np.concatenate([[0,1], np.cumsum(LPsteps[:-1],dtype=int) + shift])
        
    return Indexes


def _getBackbonePeaks(hysteresis, xyPosInd):
    
    Ncycle = len(xyPosInd)
    
    CycleIndx = xyPosInd[1:] - 1
    
    cycles = hysteresis.getCycles(CycleIndx)
    peaks = np.zeros([Ncycle,2])
    for ii in range(Ncycle - 1):
        # Set the peak, then find the local index of the maximum point
        cycles[ii].setPeaks()
        tempInd = cycles[ii].maxIndexes
        
        # Get the xy coords, the aliase over the peak indexes.
        tempxy = cycles[ii].xy
        largestPeakInd = np.argmax(tempxy[tempInd,1])
        peaks[ii] = tempxy[tempInd[largestPeakInd]]

    return np.array(peaks)





def _checkBBinput(returnPeaks, returnEnd):
    if returnPeaks == False and  returnEnd == False:
        raise ValueError('Either peaks or end points must be set.')

def _skipCycles(xyPosInd, skipStart, skipEnd, Ncycle):
    # Skip any cycles that have been specified at the start or end of the script.
    endIndex = Ncycle - skipEnd
    startIndex = skipStart
    return xyPosInd[startIndex:endIndex]
    
    
"""
TODO:
    Allow for getting the peaks, end point, or both.
"""
def getBackboneCurve(hysteresis, LPsteps = None, returnPeaks = False,  returnEnd = True, 
                     skipStart = 0, skipEnd =0, includeNegative = False):
    """
    Returns the positve backbone curve of a hysteresis.
        
    By default all reversal points on the backbone curve are returned.
    The reversal points will be the **end** points of each cycle, but for some 
    hystereses we will need to include peak points to capture the full backbone
    To capture theses peak value, we can set the "returnPeaks" flag to be true.
    
    The function will only include the positive backbone by default, however,
    the negative values can be returned using the 
    
    
    The user can output only the first cycle of each load step by specifying 
    the number of loading cycles there are at each load step.
    By default the final point (the right most point) is returned for each
    cycle.
    The user can return the peak and final point by setting returnPeaks = True.


    Parameters
    ----------
    hysteresis : Hysteresis class
        The input Hysteresis to calculate the backbone of.
    LPsteps : list, optional
        The list of the number of cycles at each load point. The default is [].
        
        For example, if the input loading is as follows: 
        
        0, 1, -1, 1, -1, 1 , -1, 2, -2, 3,-3, 3, -3
        
        Then the input should be [3, 1, 2].
        
    returnPeaks : bool, optional
        A switch that when toggled on will cause the function to return 
        the peak points of each cycle.    
    returnEnd : bool, optional
        A switch that when toggled on will cause the function to return 
        the end points of each cycle.         
    skipStart : int, optional
        The number of cycles to skip from the start of the the backbone 
        curve. The default is 0 which skips no cycles.
    skipEnd : int, optional
        The number of cycles to skip from the end of the the backbone 
        curve. The default is 0 which skips no cycles.
    includeNegative : bool, optional
        A boolean that will make the function return a negative backbone as 
        well once toggled true.
        
    Returns
    -------
    backbone : SimpleCycle
        A simpleCycle object of the output backbone curve.

    """
    
    xyPos = np.array([])
    xyPos.shape = (0,2)
    
    _checkBBinput(returnPeaks, returnEnd)    
    
    # Get a slice of the reversal indexes
    reversalIndexes = hysteresis.reversalIndexes
    xyPoints = hysteresis.xy[reversalIndexes]
    
    # Seperate out positive cycles
    diff = np.diff(xyPoints[:,0])
    xyPosInd = np.where(0 <= diff)[0] + 1
    xyPosInd = np.concatenate([[0], xyPosInd])

    # If it doesn't start at zero for some reason, make it start at zero.
    # We should always include the first index!
    if xyPosInd[0] != 0:
        xyPosInd = np.hstack((0,xyPosInd))
        
    # get the indexes of the final cycle
    if LPsteps != None: # If no value is provided
        Indexes = _LPparser(LPsteps)
        xyPosInd = xyPosInd[Indexes]
        
    Ncycle = len(xyPosInd)
       
    # skip the start/end cycles    
    xyPosInd = _skipCycles(xyPosInd, skipStart, skipEnd, Ncycle)
    
    # Get the end indexes if asked
    if returnEnd == True:
        xyPosCycleEnd = xyPoints[xyPosInd]
        xyPos = np.concatenate([xyPos, xyPosCycleEnd])
    
    # Get the postive indexes        
    if returnPeaks == True:
        xyPosPeak = _getBackbonePeaks(hysteresis, xyPosInd)
        xyPos = np.concatenate([xyPos, xyPosPeak])
    
       
    # Include the negative curve using some recursion magic
    if includeNegative:
        curve = getBackboneCurve(Hysteresis(-hysteresis.xy), LPsteps, returnPeaks,  
                         returnEnd, skipStart, skipEnd, False)
        xyNeg = -curve.xy[::-1]
        xyPos = np.concatenate((xyNeg, xyPos)) 
        
    # Remove repeated points
    xPos = xyPos[:,0]
    _, indexes = np.unique(xPos,True)
    xyPos = xyPos[indexes]   
    
    return SimpleCurve(xyPos, True)


def _averageBackbones(backBonePos,backBoneNeg):
    xPos = backBonePos.xy[:,0]
    xNeg = backBoneNeg.xy[:,0]
    
    f1 =  interp1d(backBonePos.xy[:,0], backBonePos.xy[:,1], fill_value = 'extrapolate')
    f2 =  interp1d(backBoneNeg.xy[:,0], backBoneNeg.xy[:,1], fill_value = 'extrapolate')
    
    xnet = np.concatenate((xPos, xNeg))
    x = np.unique(xnet)
    
    ypos = f1(x)
    yneg = f2(x)
    
    yavg = (ypos + yneg) / 2
    
    xy = np.column_stack([x, yavg])
    backBoneAvg = SimpleCurve(xy, True)
    
    return backBoneAvg


def getAvgBackbone(hysteresis, LPsteps = [], returnPeaks = False, returnEnd = True,
                   skipStart = 0, skipEnd =0):
    
    """
    Returns the average, positve, and negative backbone curve of a hysteresis.
        
    By default all reversal points on the backbone curve are returned.
    This will not be appropriate for some types of hysteresis where the load
    drops between cycles.
    The user can output only the first cycle of each load step by specifying 
    the number of loading cycles there are at each load step.
    By default the final point (the right most point) is returned for each
    cycle.
    The user can return the peak and final point by setting returnPeaks = True.


    Parameters
    ----------
    hysteresis : Hysteresis class
        The input Hysteresis to calculate the backbone of.
    LPsteps : list, optional
        The list of the number of cycles at each load point. The default is [].
        
        For example, if the input loading is as follows: 
        
        0, 1, -1, 1, -1, 1 , -1, 2, -2, 3,-3, 3, -3
        
        Then the input should be [3, 1, 2].
        
    returnPeaks : bool, optional
        A switch that when toggled on will cause the function to return 
        the peak points of each cycle in additon to the end points.    
    skipStart : int, optional
        The number of cycles to skip from the start of the the backbone 
        curve. The default is 0 which skips no cycles.
    skipEnd : int, optional
        The number of cycles to skip from the end of the the backbone 
        curve. The default is 0 which skips no cycles.
    Returns
    -------
    avg, pos, neg : SimpleCycle
        SimpleCycle objects of the output backbone curve.

    """
    
    
    hysNeg = Hysteresis(-hysteresis.xy)
    hysNeg.recalculateCycles_like(hysteresis)
    backBonePos = getBackboneCurve(hysteresis, LPsteps, returnPeaks, returnEnd, skipStart, skipEnd)
    backBoneNeg = getBackboneCurve(hysNeg, LPsteps, returnPeaks, returnEnd, skipStart, skipEnd)
    
    backBoneAvg = _averageBackbones(backBonePos, backBoneNeg)

    
    return backBoneAvg, backBonePos, backBoneNeg


def _linInterpolateY(xy, yinter, Index):
    
    """ Returns the x intercept of a target point y with a xy curve,
    Given the first index greater than the input intercept.
    
    """
    y1 = xy[Index - 1,1]
    y2 = xy[Index    ,1]
    x1 = xy[Index - 1,0]
    x2 = xy[Index    ,0]
    
    return linearInterpolation(y1,y2,x1,x2,yinter)

# =============================================================================
# Fit EEEP
# =============================================================================


def _get_ult_1(Pend, dUend):
    """
    Gets the ultimate point if the hystresis doesn't reach the failure load.
    """
    
    Pult  = Pend
    dUult = dUend
        
    return Pult, dUult




def _get_ult_2(Plim, xy, xyFail):
    """ 
    Gets the ultimate point if the final backbone point is lower than the 
    failure load.
    Linear interploation is used to find where the decending slope of xy
    is equal ot the failure load.
    
    """
    
    Pult  = Plim
    
    # use linear interpolation to find the intersection point.
    # Find the first index greater than the failure index
    failIndex = np.argmax(xyFail[:,1] < Plim)
    dUult     = _linInterpolateY(xyFail, Plim, failIndex)

    # add the new point, then find the index
    xy       = np.concatenate([xy, [[dUult, Pult]]])
    order    = np.argsort(xy[:,0])
    xy       = xy[order]
    endIndex = np.where(xy[:,0] == dUult)[0][0] + 1
    
    # make a new curve that ends at the failure load. 
    backbone = SimpleCurve(xy[:endIndex])
    
    return Pult, dUult, backbone

def _getIntersection(xy, Ppeak, intersection):
    # Find the elastic intercept and slope.
    Pinter  = intersection*Ppeak
    index   = np.argmin(xy[:,1] < Pinter)
    dUinter = _linInterpolateY(xy, Pinter, index)
    Ke      = Pinter / dUinter
    
    return Ke




def fitEEEP(backbone):
    """
    Fits a backbone curve with a equivalent elastic perfectly plastic curve
    using the ASTM E2126 methodology.
    The curve has equivalent area to the input backone curve.
    
    http://www.materialstandard.com/wp-content/uploads/2019/10/E2126-11.pdf

    Parameters
    ----------
    backbone : SimpleCycle
        The input SimpleCycle Object.

    Returns
    -------
    SimpleCycle
        A simpleCycle object of the backbone curve.

    """
    
    
    # There are two options - either there is a decline or there is no decline
    xy       = backbone.xy
    PpeakInd = np.argmax(xy[:,1])
    Ppeak    = xy[PpeakInd,1]
    
    # Find the final values of P and u 
    dUend, Pend = xy[-1,:] 
    xyFail      = xy[PpeakInd:,:]
    
    # ratio = Pmax / Ppeak
    Rlim = 0.8
    Plim = Rlim*Ppeak
    
    # Find the ultimate point
    if Plim < Pend: # If the final point is greater than 0.8Ppeak, use that point.
        Pult, dUult = _get_ult_1(Pend, dUend)
    else:           # If the final point is less than 0.8Ppeak, find the intercept.
        Pult, dUult, backbone = _get_ult_2(Plim, xy, xyFail)
    
    backbone.setArea()    
    Anet = backbone.getNetArea()    
    
    Ke      = _getIntersection(xy, Ppeak, 0.4)
        
    radicand = dUult**2 - 2*Anet/Ke
    
    if 0 < radicand:
        Py = Ke*(dUult - (radicand)**0.5)
    else:
        Py = 0.85*Ppeak
    
    curvexy = np.column_stack([[0, Py/Ke, dUult], [0,Py,Py]])
    curve = SimpleCurve(curvexy)
    curve.setArea()
    
    return curve






# -*- coding: utf-8 -*-
"""
Created on Fri May  7 20:15:19 2021

@author: Christian
"""

import matplotlib.pyplot as plt
from matplotlib.widgets import Button

# from matplotlib.animation import FuncAnimation
import numpy as np

def init_gui():
    fig, ax = plt.subplots()
    
    return fig, ax        



class GUIBase:
    

    def __init__(self):
        fig, ax = init_gui()
        self.fig = fig
        self.ax = ax
        pass
        
    def addNextBtns(self):
        
        self.axprev = plt.axes([0.7, 0.05, 0.1, 0.075])
        self.axnext = plt.axes([0.81, 0.05, 0.1, 0.075])
        self.bnext = Button(self.axnext, 'Next')
        self.bnext.on_clicked(self.plotNext)
        self.bprev = Button(self.axprev, 'Previous')
        self.bprev.on_clicked(self.plotPrev)
        
    def setlims(self, xlims,ylims):
        if xlims != []:
            self.ax.set_xlim(*xlims)
        if ylims != []:
            self.ax.set_xlim(*ylims)            

class CycleViewer(GUIBase):

    
    
    def __init__(self, hysteresis, showCumulative = False):
        
        super().__init__()
        
        self.showCumulative = showCumulative 
        
        
        self.hysteresis = hysteresis
        self.curves = np.array(hysteresis.Cycles)
        
        self.ind = 0
        self.Ncurves = hysteresis.NCycles
        # xyAni = getAnixy(Curve.xy, skipFrames)
        
        # add line
        self.line = self.curves[-1].plot()
        
        # a list of active curves
        # self.activeLines =  [self.line]
        
        
        # adjsut plot and add buttons
        plt.subplots_adjust(bottom=0.2)
        self.addNextBtns()
        
        plt.show()
    
    # def getActiveCurves    
    
    
    
    # def initAnimation(self):
    def updateplt(self, ind):
        # print('pressed')
        
        # xy = []
        # for ii in range(ind):
        #     xy.append(self.curves[ind].xy)
        
        # xy = np.concatenate(xy)
        xy = self.curves[ind].xy
        self.line.set_xdata(xy[:,0])
        self.line.set_ydata(xy[:,1])
        plt.draw()
        
    # def updatepltCumulative
        
    def plotNext(self, event):
        self.ind = (self.ind + 1) % self.Ncurves
        self.updateplt(self.ind)

        
    def plotPrev(self, event):
        self.ind = (self.ind - 1) % self.Ncurves
        self.updateplt(self.ind)


class CycleViewer(GUIBase):
    """
    An interactive viewer for figure to be plotted.
    
    
    Plots must use a qt backend for the GUI elements to work
    correctly. %matplotlib qt
    

    Parameters
    ----------
    hysteresis : TYPE
        The input hystesis.
    plotMethod : TYPE, optional
        A flag that toggles if single XY (0) and cumulative xy (1) is used. 
        The default is 'single'.

    Returns
    -------
    None.

    """
    plotMethods = {}
    
    def __init__(self, hysteresis, plotMethod = 0, xlims=[],ylims = []):

        super().__init__()
        self.setlims(xlims,ylims)
        
        
        # Chhose the plot method
        self.plotMethod = plotMethod 
        plotMethods = [self.getSingleXy, self.getCumXy]
        self.getPlotXY = plotMethods[plotMethod]
        
        # Store the data.
        self.hysteresis = hysteresis
        self.curves = np.array(hysteresis.cycles, dtype = object)
        # print(self.curves)
        
        # self.ind = np.array([0],dtype=int)
        self.ind = 0
        self.Ncurves = hysteresis.NCycles
        # xyAni = getAnixy(Curve.xy, skipFrames)
        
        # add initial lines and text
        self.line = self.curves[-1].plot()
        self.annot = self.ax.annotate(self.ind,(0.05,0.05), xycoords='axes fraction')        
        
        # adjsut plot and add buttons
        plt.subplots_adjust(bottom=0.2)
        self.addNextBtns()
        
        # start plot
        plt.show()
        # self.updateplt(0)   
    
    
    def getSingleXy(self, ind):
        curve = self.curves[ind]
            
        return curve.xy

    
    def getCumXy(self,ind):
        curves = self.curves[:ind+1]
        xy = np.array([])
        xy.shape = (0,2)

        for curve in curves:
            
            tempxy = curve.xy
            xy = np.concatenate([xy,tempxy])
            
        return xy
    
    def updateplt(self, ind):

        xy = self.getCumXy(ind)

        self.line.set_xdata(xy[:,0])
        self.line.set_ydata(xy[:,1])
        
        text = f'Cycle {self.ind}'

        self.annot.set_text(text)
        
        
        plt.draw()
        
        
    def plotNext(self, event):
        self.ind = (self.ind + 1) % self.Ncurves
        self.updateplt(self.ind)

        
    def plotPrev(self, event):
        self.ind = (self.ind - 1) % self.Ncurves
        self.updateplt(self.ind)

    # def plot():
    #     plt.show()

    # def CyclePlot()
    
            # def initAnimation(self):
# def updateplt(self, ind):
#     print('pressed')
#     xy = self.curves[ind].xy
#     self.line.set_xdata(xy[:,0])
#     self.line.set_ydata(xy[:,1])
#     plt.draw()
    
# def plotNext(self, event):
#     print('pressed')
#     self.ind = (self.ind + 1) % self.Ncurves
#     self.updateplt(self.ind)

    
# def plotPrev(self, event):
#     self.ind = (self.ind - 1) % self.Ncurves
#     self.updateplt(self.ind)
        
        



import hysteresis.curve as hys
import numpy as np

def makeCurve(f):
    "lamda function that creates all curves"
    xdata = np.linspace(0,4,1001)
    y = f(xdata)
    xy = np.column_stack([xdata, y])
    Curve = hys.Curve(xy)   
    return Curve    

def function1(x):
    return x + 2

def function2(x):
    return x**3 - 3*x**2 + 3

def function3(x):
    return 3 + (np.e)**x  
        
def test_Curve1_Area():
    Curve = makeCurve(function1)
    Curve.setSlope()
    slope = Curve.slope
    
    test1 = (abs(slope[0] - 1) < 0.0001)
    test2 = (abs(slope[-1] - 1) < 0.0001)
    test3 = (abs(slope[50] - 1) < 0.0001)

    assert np.all([test1,test2,test3])

def test_Curve2_Area():
    Curve = makeCurve(function2)
    Curve.setSlope()
    slope = Curve.slope
    
    test1 = (abs(slope[0] - 0) < 0.05)
    test2 = (abs(slope[-1] - 24) < 0.05)
    test3 = (abs(slope[200] - (3*0.8**2 - 6*0.8)) < 0.0001)

    assert np.all([test1,test2,test3])

def test_Curve3_Area():
    Curve = makeCurve(function3)
    Curve.setSlope()
    slope = Curve.slope

    test1 = (abs(slope[0] - 1) < 0.01)
    test2 = (abs(slope[-1] - (np.e)**4) < 0.2)
    test3 = (abs(slope[200] - ((np.e)**0.8)) < 0.0001)

    assert np.all([test1,test2,test3])


# out = test_Curve3_Area()
    





import hysteresis as hys
import numpy as np
import matplotlib.pyplot as plt



x = np.linspace(0,1,101)
xy = np.column_stack([x,x])

hys.Hysteresis(xy)


out = hys.concatenateHys((xy,-xy))


halfCycle = True

loadProtocol = [0.001547,  0.00221, 0.003315, 0.00442, 0.00663, 0.00884, 0.01326,
                0.01768,   0.02652, 0.03978,  0.05967, 0.0884,  0.1326,  0.1768, 0.221, 
                0.3315,    0.4199,  0.4862]


Nrepeat     = 3*np.ones_like(loadProtocol, int)


protocol = hys.createProtocol(loadProtocol, Nrepeat) 
Npoint = len(protocol)
Npoint = np.arange(Npoint)
    
xy = np.column_stack((Npoint, protocol))

hys.SimpleCurve(xy)

# def _expandProtcol(protcol, halfCycle, Nrepeats):
    
#     # Check if Nrepeats matches protcol
    
    
#     if halfCycle:
        
#     output = [None]*np.sum()
#     for item in loadProtocol:
        


# outputProtcol


import numpy as np
import hysteresis as hys


def test_comapre_basic():
    t = np.linspace(0,4,1000)*np.pi
    x = np.sin(t)
    y = np.cos(t)*t
    
    xy = np.column_stack([x,y])
    
    
    
    myHys = hys.Hysteresis(xy)
    myHys.plot(showReversals = True)
    smallHys = hys.resample(myHys, 10)
    # samllHys.plot()
    
    out = hys.compareHys(smallHys,myHys)
    
    assert out[0] == 0




# t = np.linspace(0,4,1000)*np.pi
# x = np.sin(t)
# y = np.cos(t)*t

# xy = np.column_stack([x,y])



# myHys = hys.Hysteresis(xy)
# myHys.plot(plotCycles = True)
# smallHys = hys.reSample(myHys, 10) 
# smallHys.xy[:,1] = smallHys.xy[:,1] / 4
# smallHys = hys.Hysteresis(smallHys.xy)


# smallHys.plot()
# # smallHys.xy[:,1] = smallHys.xy[:,1]/2


# out = hys.CompareHys(smallHys,myHys)


# -*- coding: utf-8 -*-
"""
Created on Fri Aug 21 23:47:36 2020

@author: Christian
"""
import hysteresis as hys
from hysteresis.protocol import getReturnCycle
import numpy as np
import matplotlib.pyplot as plt
        

def test_getReturnCycle():
    """ Tests if the getReturnCycle function"""
    t = np.linspace(0,1,101)
    x1 = 1 - 1.5*t
    y1 =  (3)*x1**2 - 1
    x2 = x1[-1] + t*2
    y2 = y1[-1] + t*4
    
    
    TestCycle1 = hys.SimpleCurve(np.column_stack([x1,y1]))
    TestCycle2 = hys.SimpleCurve(np.column_stack([x2,y2]))
    TestCycle3 = getReturnCycle(TestCycle1, TestCycle2)
    
    xySolution = np.zeros([76,2])
    xySolution[:75, :] = TestCycle2.xy[:75, :]
    xySolution[-1, :] = TestCycle1.xy[0, :]
    
    assert(np.all(xySolution == TestCycle3.xy))
    